from chainlogger.parser_class import ParserClass
import hashlib


class Hasher:
    """
        A class to hash the values, add salt and verify with given values.

        ...

        Attributes
        ----------
        salt : str
            a salt value to randomize the hash

        Methods
        -------
        create(string)

        add0x(string)

        hash_with_salt(string)

        verify_input(string, string)
    """

    def __init__(self, salt):
        """
        Initialize the given class.
            Parameters:
                salt (string): a salt value to randomize the hash
        """
        self.salt = salt

    @staticmethod
    def create(salt):
        """
        Returns the initialized class.
            Parameters:
                salt (string): a salt value to randomize the hash
        Returns:
                Hasher (class): A hasher class
        """
        return Hasher(salt)

    @staticmethod
    def add0x(input_x):
        """
        Returns the "0x" added version of the input.
            Parameters:
                input_x (string): a data
            Returns:
                value (string): "0x" merged with a data
        """
        return "0x" + input_x

    def hash_with_salt(self, input_x):
        """
        Returns the hashed with sha256 version of the data.
            Parameters:
                input_x (string): a data
            Returns:
                value (string): hashed and salted version of the data
        """
        hashed_data = hashlib.sha256((input_x + self.salt).encode()).hexdigest()
        return hashed_data

    def verify_input(self, hashed_input, raw_data):
        """
        Takes the raw input and hashed_input, and makes a comparison.
            Parameters:
                hashed_input (string): a hashed data
                raw_data (string): a raw data
            Returns:
                value (boolean): are they the same or not
        """
        to_check = Hasher.add0x(self.hash_with_salt(ParserClass.json_encode(element=raw_data)))

        if hashed_input == to_check:
            return True
        else:
            return False



import json


class ParserClass:
    """
        A parser class to interchange between json objects and python dicts.
        ...
        Methods
        -------
        json_encode(dict)

        json_decode(string)
    """

    @staticmethod
    def json_encode(element):
        """
        Returns the json representation of given element.
            Parameters:
                element (dict): A python dict
        Returns:
                value (str): A string representation of the json object
        """
        return json.dumps(element)

    @staticmethod
    def json_decode(element):
        """
        Returns the dict representation of given json object.
            Parameters:
                element (string): A json object
        Returns:
                value (str): A dict representation of the json object
        """
        return json.loads(element)



from web3 import Web3
from web3.middleware import geth_poa_middleware


class Contract:
    """
    A class to initialize contracts, gets the provider and abi information about the contract.
    Initializes web3 instances.
    ...
    Methods
    -------
    set_provider(string)

    set_abi(string)

    set_eth_signer(string)

    set_contract(string)

    set_send_contract(string)

    set_web3(string)

    get_contract()

    get_send_contract()

    get_web3()
    """
    provider = None
    abi = None
    eth_signer = None
    contract = None
    send_contract = None
    web3 = None

    @staticmethod
    def check_connection_urls(provider, abi):
        """
        Checks all the variables exists
        Parameters:
            provider (string): provider url
            abi (string): abi
        """
        if provider is None:
            raise Exception("You must set the provider first")
        if abi is None:
            raise Exception("You must set the abi of the contract first")

    def set_provider(self, provider):
        """
        Sets the provider
        Parameters:
            provider (string): provider url
        """
        self.provider = provider

    def set_abi(self, abi):
        """
        Sets the abi
        Parameters:
            abi (string): contract abi
        """
        self.abi = abi

    def set_eth_signer(self, eth_signer):
        """
        Sets the eth signer
        Parameters:
            eth_signer (string): eth signer url
        """
        self.eth_signer = eth_signer

    def set_contract(self, at_address):
        """
        Sets the deployed contract
        Parameters:
            at_address (string): deployed contract address
        """
        self.check_connection_urls(self.provider, self.abi)
        web3_instance = Web3(Web3.HTTPProvider(self.provider))
        if not web3_instance.isConnected():
            raise ConnectionError
        web3_instance.middleware_onion.inject(geth_poa_middleware, layer=0)
        self.contract = web3_instance.eth.contract(address=web3_instance.toChecksumAddress(at_address), abi=self.abi)

    def set_send_contract(self, send_contract_address):
        """
        Sets the transaction sending contract using EthSigner provider
        Parameters:
            send_contract_address (string): sender contract address
        """
        self.check_connection_urls(self.provider, self.abi)
        eth_signer_instance = Web3(Web3.HTTPProvider(self.eth_signer))
        if not eth_signer_instance.isConnected():
            raise ConnectionError
        eth_signer_instance.middleware_onion.inject(geth_poa_middleware, layer=0)
        self.send_contract = eth_signer_instance.eth.contract(
            address=eth_signer_instance.toChecksumAddress(send_contract_address),
            abi=self.abi)

    def set_web3(self):
        """
        Sets the Web3 object
        """
        if self.provider is None:
            raise Exception("You must set the provider first")
        self.web3 = Web3(Web3.HTTPProvider(self.provider))
        if not self.web3.isConnected():
            raise ConnectionError

    def get_contract(self):
        """
            Gets the contract
        """
        return self.contract

    def get_send_contract(self):
        """
            Gets the contract
        """
        return self.send_contract

    def get_web3(self):
        """
            Gets the Web3 object
        """
        return self.web3



from chainlogger.hasher import Hasher
from chainlogger.parser_class import ParserClass
from chainlogger.contract import Contract


class Logger(Contract):
    """
        A class to contain logging logics.
        ...
        Methods
        -------
        create()

        set_gas(string)

        set_gas_price(string)

        set_account(string)

        set_salt(string)

        register_vendor()

        register_project()

        register_log(int, string)

        get_transaction_receipt(string)

        get_batch_transaction_receipt(list)

        parse_data(string)

        parse_batch_data(list)

        verify_data(string, string, string)
    """

    account = None
    salt = "CHANGE_ME"
    gas = "0x3d090"
    gas_price = "0x174876e800"

    @staticmethod
    def create():
        """
        Static function that returns the Logger instance.
        Returns:
            value (object): Logger instance
        """
        return Logger()

    def set_gas(self, gas):
        """
        Sets the gas.
        """
        self.gas = gas

    def set_gas_price(self, gas_price):
        """
        Sets the gas price.
        """
        self.gas_price = gas_price

    def set_account(self, account):
        """
        Sets the user's account from EthSigner.
        """
        self.account = account

    def set_salt(self, salt):
        """
        Sets the salt.
        """
        self.salt = salt

    def register_vendor(self):
        """
        Calls the registerVendor method on the omChain
        Returns the transaction hash or throws error
        Takes gasPrice and gas as hexadecimal
        Returns:
            value (string): transaction hash
        """
        parameters = {
            'from': self.account,
            'gasPrice': self.gas_price,
            'gas': self.gas
        }
        try:
            tx_hash = self.send_contract.functions.registerVendor().transact(parameters)
        except Exception as e:
            print(str(e))
            raise

        return tx_hash

    def register_project(self):
        """
        Registers a project
        Returns:
            value (string): transaction hash
        """
        parameters = {
            'from': self.account,
            'gasPrice': self.gas_price,
            'gas': self.gas
        }
        try:
            tx_hash = self.send_contract.functions.registerProject().transact(parameters)
        except Exception as e:
            print(str(e))
            raise

        return tx_hash

    def register_log(self, project_id, raw_data):
        """
        Registers a log
        Returns:
            value (string): transaction hash
        """
        parameters = {
            'from': self.account,
            'gasPrice': self.gas_price,
            'gas': self.gas
        }
        hasher = Hasher(self.salt)
        to_chain_data = ParserClass.json_encode(raw_data)
        to_chain_data = hasher.hash_with_salt(to_chain_data)
        to_chain_data = Hasher.add0x(to_chain_data)

        try:
            tx_hash = self.send_contract.functions.registerLog(project_id, to_chain_data).transact(parameters)
        except Exception as e:
            print(str(e))
            raise

        tx_hash = self.web3.toHex(tx_hash)
        return tx_hash

    def get_transaction_receipt(self, tx_id):
        """
        Gets transaction receipt with events
        Returns:
            value (dict): data dictionary
        """
        receipt = self.web3.eth.get_transaction_receipt(tx_id)
        result = self.contract.events.LogRegistered().processReceipt(receipt)
        if len(result) == 0:
            raise
        event_log_data_ls = []
        for log in result:
            data_dict = {
                "_data": self.web3.toHex(log["args"]["_data"]),
                "_projectId": log["args"]["_projectId"],
                "_projectLogCounter": log["args"]["_projectLogCounter"],
                "_vendorAddress": log["args"]["_vendorAddress"]
            }
            event_log_data_ls.append({
                "transactionHash": self.web3.toHex(log['transactionHash']),
                "blockHash": self.web3.toHex(log['blockHash']),
                "blockNumber": log['blockNumber'],
                "data": data_dict
            })

        return event_log_data_ls[0]

    def get_batch_transaction_receipt(self, tx_ids):
        """
        Get Batch Transaction Receipts
        Returns:
            value (list): data dictionary list
        """
        return_array = []
        for tx_id in tx_ids:
            return_array.append(self.get_transaction_receipt(tx_id))

        return return_array

    @staticmethod
    def parse_data(block_data):
        """
        Parses the transaction receipt's data column
        to return string instead of hexadecimal
        Returns:
            value (list): parsed data
        """
        if not isinstance(block_data, list):
            return block_data
        return_data = []
        for block in block_data:
            return_data.append(block)

        return return_data

    def parse_batch_data(self, block_data_array):
        """
        Parse Batch Block Data
        Returns:
            value (list): parsed data list
        """
        return_data = []
        for block_data in block_data_array:
            return_data.append(self.parse_data(block_data))

        return return_data

    def verify_data(self, hashed_data, raw_input, salt=None):
        """
        Verify integrity of blockchain record
        Returns:
            value (bool): is verify value
        """
        if salt:
            hasher = Hasher.create(salt)
        else:
            hasher = Hasher.create(self.salt)

        to_chain_data = ParserClass.json_encode(raw_input)
        to_chain_data = hasher.hash_with_salt(to_chain_data)
        to_chain_data = Hasher.add0x(to_chain_data)

        if hashed_data == to_chain_data:
            return True
        else:
            return False



from setuptools import setup, find_packages
import os

here = os.path.abspath(os.path.dirname(__file__))

VERSION = '0.0.1'
DESCRIPTION = 'Log data into blockchain'
LONG_DESCRIPTION = 'A package that register a project, vendor and a log into a selected blockchain.'

setup(
    name="chainlogger",
    version=VERSION,
    author="MERT YILMAZ",
    author_email="<mert@omlira.com>",
    description=DESCRIPTION,
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    packages=find_packages(),
    install_requires=['web3'],
    keywords=['python', 'blockchain', 'logging'],
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3"
    ]
)
